
#include <stdlib.h>
#include <utils/commandline.h>
#include <gridfastslam/gridslamprocessor.h>
#include <utils/orientedboundingbox.h>
#include <configfile/configfile.h>
#include <libplayerc++/playerc++.h>
#include <sensor/sensor_base/sensorreading.h>
#include <sensor/sensor_range/rangesensor.h>
#include <sensor/sensor_range/rangereading.h>
//#include </usr/local/include/player-2.0/libplayerc++/playerc++.h>
#include <iostream>
#include <fstream>
#include <time.h>

#include <QtGui/qpainter.h>
#include <QtGui/qpixmap.h>
#include <QtGui/qwidget.h>
#include <QtGui/qwmatrix.h>
#include <Qt/qtextstream.h>
#include <QtGui/qimage.h>

#include <vector>

//#include "motion.h"
#include "laserFusion.h"

//#define PI 3.1415

using namespace GMapping;
using namespace std;

void showHelp()
{
    using namespace std;

    cout << "SLAM program+Motion Control" << endl;
    cout << "Adapted by Yi Li, Konstantinos Bitsakos" << endl;
    cout << "Usage: gfs-erratic-motion <HostIP> <PortNumber> <TiltStep> <TiltPositions>" << endl;
    cout << "<HostIP>        : The IP address of the computer hosting the player" << endl;
    cout << "<PortNumber>    : The number of the port listening for incoming player connections" << endl;
    cout << "<TiltStep>      : The tilt step (default:5 degrees)" << endl;
    cout << "<TiltPositions> : The number of tilt positions (default:11)" << endl; 
    return;
}

bool writeImageToFile(char *a_imFilename, GMapping::Map<double, GMapping::DoubleArray2D, false> &a_map, OrientedPoint * ap_robotPose)
{
    QImage image(a_map.getMapSizeY(),a_map.getMapSizeX(), QImage::Format_RGB32);
    image.fill(qRgb(255,255,255));

    for(int x=0; x<image.width(); x++)
    	for(int y=0; y<image.height(); y++)
    	{
            double v=a_map.cell(GMapping::IntPoint(x,y));
    		if (v>=0)
    		{
    			int grayValue=255-(int)(255.*v);
    			image.setPixel(x,y,qRgb(grayValue, 0, 0));
    		}
    	}
    if (ap_robotPose)
    {
            image.setPixel((int)(ap_robotPose->x), (int)(ap_robotPose->y),qRgb(0,255,0));
            cout << "Writing to image file:Pose(" << ap_robotPose->x << "," << ap_robotPose->y << ")" << endl;
            for (int i=1; i<4; i++)
            {
                image.setPixel((int)(ap_robotPose->x+i*cos(ap_robotPose->theta)),
                               (int)(ap_robotPose->y+i*sin(ap_robotPose->theta)),
                               qRgb(0,0,255) ); 
            }    
    }    

	return image.save(a_imFilename,"PNG");
}

bool dumpPoseToFile(char *a_filename, IntPoint a_position, double theta, double a_laserXpos, double a_laserYpos, double a_laserTheta)
{
    // Open a file handle
    ofstream    outFile(a_filename, ios::out);
    if (!outFile)
    {
        cout << "ERROR opening input file" << endl;
        return false;
    }
    outFile << a_position.x << " " << a_position.y << " " << theta << endl;
    outFile << a_laserXpos << " " << a_laserYpos << " " << a_laserTheta << endl;
    outFile.close();
    return true;
}

void dumpMapToFile(const Map<double, DoubleArray2D, false> &a_map, char *a_filename,double a_worldSizeX, double a_worldSizeY)
{
    // Open a file handle
    ofstream    outFile(a_filename, ios::out);
    if (!outFile)
    {
        cout << "ERROR opening input file" << endl;
        return;
    }
    outFile << a_map.getMapSizeY() << " " << a_map.getMapSizeX() << " " << 
    a_map.getWorldSizeY() << " " << a_map.getWorldSizeX() 
//    a_worldSizeY << " " << a_worldSizeX 
    << " " << a_map.getDelta() << endl;

    for (int x=0; x<a_map.getMapSizeX(); x++)
    {
        for (int y=0; y<a_map.getMapSizeY(); y++)
        {
            outFile << a_map.cell(IntPoint(x,y)) << " ";
        }
        outFile << endl;
	}			
    outFile.close();
    return;
}

int main(int argc, const char * const * argv) {

    using namespace std;

    // Player connection parameters
    int portNum=6665;

    // Map parameters
	double xmin=-10.;
	double ymin=-10.;
	double xmax=10.;
	double ymax=10.;
	double delta=0.05; //0.05; // TODO: Check why the code fails for all values other than 0.05 (and multiples)

	//scan matching parameters
	double sigma=0.05;
	double maxrange=4;
	double maxUrange=4;
//	double regscore=1e4;
	double lstep=.05;
	double astep=.05;
	int kernelSize=1;
	int iterations=5;
//	double critscore=0.;
//	double maxMove=1.;
	double lsigma=.075;
	double ogain=3;
	int lskip=0;

	//motion model parameters
	double srr=0.01, srt=0.01, str=0.01, stt=0.01;
	//particle parameters
	int particles=20;//30;


	//gfs parameters
	double angularUpdate=15*3.14/180;//0.5;
	double linearUpdate=0.25;
	double resampleThreshold=0.5;
	bool generateMap=true;

    // tilt parameters
    double  tiltStep=5*PI/180;
    unsigned int  tiltIterations=11;
//    double tiltOffset=32*PI/180;

    // Motion parameters
//    unsigned int uSleepTime=10000; // Sleep for 10 milisecond between iterations
    
    // First Process the command-line arguments
    if (argc<2) { 
        showHelp();
        return 0;
    }
    if (argc>=3) portNum=atoi(argv[2]);
    cout << "1" << endl;
    if (argc>=4) tiltStep=atof(argv[3])*PI/180; 
    cout << "2" << endl;
    
    if (argc>=5) tiltIterations=atoi(argv[4]);
    cout << "3" << endl;

    cout << argv[1] << ":" << portNum << "\tilt Step:"<< tiltStep*180/PI << " Tilt Positions:" << tiltIterations << endl;

	try
		{
/*		//CREATION
		using namespace PlayerCc;

		PlayerClient robot(argv[1], portNum);
		Position2dProxy pp(&robot, 0);
        PtzProxy ptz(&robot, 0);
		LaserProxy lp(&robot, 0);
		std::cout << robot << std::endl;
        robot.SetDataMode(PLAYER_DATAMODE_PULL); // Get data on demand. Ensures that the most up-to-date data are received
        robot.SetReplaceRule(true,-1,-1,-1); // Setup a replace rule, so the queues are not overflowed. Check "playerclient.h" for details
		pp.SetMotorEnable(true);
*/
        // Create the Laser Fusion class
        LaserFusion lf(argv[1], portNum, tiltStep, tiltIterations);
        cout << "Laser Fusion initialized" << endl;      
        

        // Create the SLAM processor class
		GridSlamProcessor* processor=new GridSlamProcessor;
//		robot.Read();

/*
        // Set the tilt unit to the horizontal position
        ptz.SetCam( (double)tiltOffset, 0, 0);

		//SENSOR MAP 
		double res= lp.GetScanRes();
		int count = lp.GetCount();

		int loop = 0;
		while( res == 0 )
		{
			robot.Read();
//			std::cout << count++ << std::endl;
			res= lp.GetScanRes();
			count = lp.GetCount();
			loop++;
			if( loop > 10 )
				return -1;
		}

		assert(res>0);
*/
		string sensorName="FLASER";

//        lf.updateMeasurements(); // Get new laser measurements
        unsigned int    count=lf.m_laserScanCount; // The number of measurements per scan
        double          res=lf.m_laserRes; // The resolution of the scan
                
//		OrientedPoint rpose(pp.GetXPos(), pp.GetYPos(), pp.GetYaw());
//		OrientedPoint lpose(pp.GetXPos(), pp.GetYPos(), pp.GetYaw());
//		OrientedPoint dp=absoluteDifference(lpose, rpose);
		RangeSensor*m_frontLaser=new RangeSensor(sensorName,count, res, OrientedPoint(0,0,0), 0,maxUrange);
		m_frontLaser->updateBeamsLookup();
		SensorMap m_sensorMap;
		m_sensorMap.insert(make_pair(sensorName, m_frontLaser));
			
		processor->setSensorMap(m_sensorMap);

		//set the command line parameters
		processor->setMatchingParameters(maxUrange, maxrange, sigma, kernelSize, lstep, astep, iterations, lsigma, ogain, lskip);
		processor->setMotionModelParameters(srr, srt, str, stt);
		processor->setUpdateDistances(linearUpdate, angularUpdate, resampleThreshold);
		processor->setgenerateMap(generateMap);
		//OrientedPoint initialPose((xmin+xmax)/2, (ymin+ymax)/2, 0);
        OrientedPoint initialPose(lf.m_xPos, lf.m_yPos, lf.m_theta); 

		//INITIALIZATION
		processor->init(particles, xmin, ymin, xmax, ymax, delta, initialPose);

//		bool running=true;

		GridSlamProcessor* ap, *copy=processor->clone();
		ap=processor; processor=copy; copy=ap;

		//this is the CORE LOOP;	
		// go into read-think-act loop
// 		int idx = 0;
		//srand(time(NULL));
        //pp.SetMotorEnable(true);
        //pp.SetSpeed(0.2, 0);
		for(;;)
		{

/*
			// this blocks until new data comes; 10Hz by default
			robot.Read();

            uint count = lp.GetCount();
*/
            // Fuse the laser measurements for different tilt angles
            lf.updateMeasurements();
            
			/////////////////////////////
			/// get readings and set slam
			/////////////////////////////
			const RangeSensor * rs=m_frontLaser;
			RangeReading reading(rs, time(NULL));
			reading.resize(rs->beams().size());
			for (unsigned int i=0; i< (unsigned int)count; i++)
			{
//				reading[i]=(double)lp[i];
				reading[i]=(double)lf.m_fusedLaserScan[count-i];
			}
			reading.setPose(OrientedPoint(lf.m_xPos, lf.m_yPos, lf.m_theta));


			bool processed=processor->processScan(reading);

			//this returns true when the algorithm effectively processes (the traveled path since the last processing is over a given threshold)
			if (processed){
                // SYNCHRONIZE WITH THE MOTION CONTROL MODULE
                // Create a file called "slamProcessing.dat"
                ofstream    outFile("slamProcessing.dat", ios::out);
                outFile.close();
                
				//for searching for the BEST PARTICLE INDEX
				unsigned int best_idx=processor->getBestParticleIndex();

				//then if you want to access the BEST MAP,
				//of course by copying it in a plain structure 
				//at this point mymap is yours. Can do what you want.
				Map<double, DoubleArray2D, false>* mymap = processor->getParticles()[best_idx].map.toDoubleMap();
//                cout << "World size:(" << mymap->getWorldSizeX() << "," << mymap->getWorldSizeY() << ")\t";
//                cout << "Map size:(" << mymap->getMapSizeX() << "," << mymap->getMapSizeY() << ")" << endl;
                // Find the pose of the robot                
                OrientedPoint robot_pose=(processor->getParticles()[best_idx]).pose;
//                cout << "Robot Pose:(" << robot_pose.x << "," << robot_pose.y << "," << robot_pose.theta << ")"<< endl;

                //dump the map somewhere.......
                dumpMapToFile(*mymap, "occupancy.dat", xmax-xmin, ymax-ymin);
                // Write the map to an image file	
                IntPoint int_pose(mymap->world2map(Point(robot_pose.x,robot_pose.y)));
//                writeImageToFile("occupancy.png", *mymap, &OrientedPoint(int_pose.x, int_pose.y, robot_pose.theta));
                // Write the pose to another file
//        		inline IntPoint world2map(const Point& p) const;
//                bool dumpPoseToFile(char *a_filename, IntPoint a_position, double theta)
                dumpPoseToFile("pose.dat", mymap->world2map(Point(robot_pose.x,robot_pose.y)), robot_pose.theta, lf.m_xPos, lf.m_yPos, lf.m_theta );
/*				double scale = 50;
//				QImage image(int(round((ymax-ymin)*scale*.95)),int(round((xmax-xmin))*scale*.95), QImage::Format_RGB32);
				QImage image(mymap->getMapSizeY(),mymap->getMapSizeX(), QImage::Format_RGB32);
				image.fill(qRgb(255,255,255));

				bool bHasSthToDraw = false;
				for(int x=0; x<image.width(); x++)
					for(int y=0; y<image.height(); y++)
					{
//						IntPoint tempC = IntPoint(x-image.width()/2, y-image.height()/2);
//						Point p=Point(tempC.x/scale, -tempC.y/scale);
//                        Point p=Point(x, y);

//						double v=mymap->cell(p);
                        double v=mymap->cell(IntPoint(x,y));
						if (v>=0)
						{
							int grayValue=255-(int)(255.*v);
							image.setPixel(x,y,qRgb(grayValue, 0, 0));
							bHasSthToDraw = true;
						}
					}

				if(bHasSthToDraw)
				{
					char name[100];
					sprintf(name,"map.png"); // , idx++
					bool rv=image.save(name,"PNG");
				}
*/
				GridSlamProcessor* newProcessor=processor->clone();
				delete processor;
				processor=newProcessor;
			} // end of if(processed)
//            usleep(uSleepTime); 
		}
	}
	catch (PlayerCc::PlayerError e)
	{
		std::cerr << e << std::endl;
		return -1;
	}
	return 0;
}

