//#include <stdafx.h>
//#define  _USE_MATH_DEFINES
#include "laserFusion.h"

using namespace std;

bool LaserFusion::Quantize( const vector<PT3D>& pts, vector<double>& OneScan )
{
	if( OneScan.size() != NUM )
	{
		OneScan.clear();
		for( int i=0; i<NUM; i++ )
			OneScan.push_back( 4.095 );
	}
	
	for( vector<PT3D>::const_iterator it=pts.begin(); it!=pts.end(); it++ )
	{
		PT3D pt = *it;
		
		if( pt.z < 0.05 )
			continue;

		double w = atan2( pt.y, pt.x );
		double r = sqrt( pt.x*pt.x + pt.y*pt.y );

		int n = (int)floor( w/DEV ) + (NUM-1)/2;

		if( n==-1 )
			OneScan.at(0) = min( r, OneScan.at(0) );
		else if( n == NUM-1 )
			OneScan.at(n) = min( r, OneScan.at(n) );
		else if( n>-1 && n<NUM-1 )
		{
			OneScan.at(n  ) = min( r, OneScan.at(n  ) );
			OneScan.at(n+1) = min( r, OneScan.at(n+1) );
		}
	}

	return true;
}

bool LaserFusion::TransformLaserLocalToGlobal( const vector<PT3D>& ptsLocal, double Tx, double Ty, double theta, vector<PT3D>& ptsGlobal )
{
	if( ptsLocal.begin() == ptsGlobal.begin() )
		return false;

	ptsGlobal.clear();

	for( vector<PT3D>::const_iterator it=ptsLocal.begin(); it!=ptsLocal.end(); it++ )
	{
		PT3D pt0 = *it;
		PT3D pt1;
		pt1.y =  pt0.y * cos(theta) + pt0.x * sin(theta) + Ty;
		pt1.x = -pt0.y * sin(theta) + pt0.x * cos(theta) + Tx;
		pt1.z = pt0.z;

		ptsGlobal.push_back( pt1 );
	}

	return true;
}

bool LaserFusion::TransformRadialTo3D( double fAngle, vector<PT3D>& pts ) //vector<double> OneScan,
{

	float H = 0.2091f;
	float h = 0.0890f;
	float a = 0.5514f;
	float b = 2.1199f*M_PI/180.0f;

	size_t sz = mp_lp->GetCount();
	pts.clear();

	double phi = ( fAngle * a + b ); //* (double)M_PI/180.0f;
	for( int w=0; w<(int)sz; w++ )
	{
		double the = (w-(int)(sz/2)) / (double)(sz-(int)(sz/2)-1) * VIW/2.0f;
		
		double l = (*mp_lp)[w];//OneScan.at(w);
        
        if( l >= INF )
			continue;
    
        double sint = sin(the);
        double cost = cos(the);
        double sinp = sin(phi);
        double cosp = cos(phi);

        double yy = 0;
        double zz = h * cosp + H;
        double xx = h * sinp;

		PT3D pt;
        pt.x = xx + l * cost * cosp;
        pt.y = yy + l * sint;
        pt.z = zz - l * cost * sinp;

		pts.push_back( pt );
	}

	return true;
}

// Constructor
/*LaserFusion::LaserFusion(PlayerCc::PlayerClient *ap_robot, PlayerCc::Position2dProxy *ap_pp, PlayerCc::PtzProxy *ap_ptz, PlayerCc::LaserProxy *ap_lp)
{
        // Initialize private variables
        m_uSleepTime=1000000;
        m_iteration=1;
        m_filename="temp_";
		m_laserScanCount=NUM;

        // Copy the points to the member variables
        mp_robot=ap_robot;
        mp_pp=ap_pp;
        mp_ptz=ap_ptz;
        mp_lp=ap_lp;
        
        // Initialize the laser sensor
        (*mp_robot).Read();
		double res= (*mp_lp).GetScanRes();
        cout << "Scan Resolution:"<< res << endl;
		int count = (*mp_lp).GetCount();

		int loop = 0;
		while( res == 0 )
		{
			(*mp_robot).Read();
			std::cout << count++ << std::endl;
			res= (*mp_lp).GetScanRes();
			count = (*mp_lp).GetCount();
			loop++;
		}
		
		// Allocate the memory for the laser scans
//		m_fusedLaserScan.vector();
		
        // Set up the tilt angles
        for (int i=0; i<5; i++)
            m_tiltAngles.push_back((-20+i*10)*PI/180);
       
        updateMeasurements();
}
*/

// Constructor
LaserFusion::LaserFusion(const char *ip_address, int port_num, double tilt_step, unsigned int tilt_positions)
{
        using namespace PlayerCc;
        // Create the necassary structures
	mp_robot=new PlayerClient(ip_address, port_num);
	mp_pp=new Position2dProxy(mp_robot, 0);
        mp_ptz=new PtzProxy(mp_robot, 0);
	mp_lp=new LaserProxy(mp_robot, 0);
        mp_robot->SetDataMode(PLAYER_DATAMODE_PULL); // Get data on demand. Ensures that the most up-to-date data are received
        mp_robot->SetReplaceRule(true,-1,-1,-1); // Setup a replace rule, so the queues are not overflowed. Check "playerclient.h" for details
		mp_pp->SetMotorEnable(true);

        // Initialize private variables
        m_uSleepTime=100000; // 0.1 sec 
		m_laserScanCount=NUM;
        m_horizontalOffset=32*PI/180;        
                
        // Initialize the laser sensor
        mp_robot->Read();
 //       sleep(2);
		double res= mp_lp->GetScanRes();
        cout << "Scan Resolution:"<< res << endl;
		int count = mp_lp->GetCount();

		int loop = 0;
		while( res == 0 )
		{
			mp_robot->Read();
			std::cout << count++ << std::endl;
			res= mp_lp->GetScanRes();
			count = mp_lp->GetCount();
			loop++;
		}
		
        // Find the maximum(minimum) angle
        double max_angle=((int)(tilt_positions/2))*tilt_step;
        for (double cur_angle=-max_angle; cur_angle<=max_angle; cur_angle+=tilt_step)
            m_tiltAngles.push_back(cur_angle);
		
        // Set up the tilt angles
//        for (int i=0; i<5; i++)
//            m_tiltAngles.push_back((-20+i*10)*PI/180);
//            m_tiltAngles.push_back(0.);
       
        updateMeasurements();
		m_laserRes= mp_lp->GetScanRes();
		m_laserScanCount = mp_lp->GetCount();
}

void LaserFusion::updateMeasurements()
{
    double initlXPos=-10., initlYPos=-10., initlTheta=-10.;
    vector<double>          tempFusedScan;

    // Create the file handle for the laser scan file and for the laser pose file



        {
	using namespace PlayerCc;
    for (vector<double>::const_iterator iter=m_tiltAngles.begin(); iter!=m_tiltAngles.end(); iter++)
    {
        vector<PT3D> curScan;
        vector<PT3D> curScanGl;
        double curwXPos=0., curwYPos=0., curwTheta=0.;
        double curlXPos=0., curlYPos=0., curlTheta=0.;
               
        // Move the tilt unit 
        mp_ptz->SetCam((double)*iter+m_horizontalOffset,0,0);
        usleep(m_uSleepTime);
		mp_robot->Read();
		
        // Get the position
        curwXPos= mp_pp->GetXPos();
        curwYPos= mp_pp->GetYPos();
        curwTheta=mp_pp->GetYaw();
               
        // Transform the coordinates from the center of the wheels to the laser sensor
        wheelsToLaserCoordinates(curwXPos, curwYPos, curwTheta, curlXPos, curlYPos, curlTheta);

        if (iter==m_tiltAngles.begin())
        {
            initlXPos=curlXPos; initlYPos=curlYPos; initlTheta=curlTheta;
        }
        
        // Get the laser data and transform them to X,Y,Z coordinates  
        TransformRadialTo3D( *iter, curScan );
                        
        // Transform the laser data to the original coordinate system
        TransformLaserLocalToGlobal( curScan, curlXPos-initlXPos, curlYPos-initlYPos, curlTheta-initlTheta, curScanGl );
        
        // Interpolate the data to create the fused measurements
        Quantize( curScanGl, tempFusedScan );
    }
    m_fusedLaserScan.clear();

        }

    ofstream    lscan("laserScan.dat");


    lscan << m_laserScanCount << endl;
    // Copy the fused measurements to the member variable
    for (vector<double>::const_iterator iter=tempFusedScan.begin(); iter!=tempFusedScan.end(); iter++)
    {
        m_fusedLaserScan.push_back(*iter);
        // Copy the Fused Scan to the file
        lscan << (*iter) << " " ;
    }

    lscan<<endl;
    lscan.close();
    ofstream    lpose("laserPose.dat");    
    // Copy the position measurements to the member variables
    m_xPos=initlXPos; m_yPos=initlYPos; m_theta=initlTheta;
    // Copy the initial laser position to a file
    lpose << initlXPos << " " << initlYPos << " " << initlTheta << endl;
    // Close both files
    lpose.close();

}

void LaserFusion::wheelsToLaserCoordinates(double curwXPos, double curwYPos, double curwTheta, double &curlXPos, double &curlYPos, double &curlTheta)
{
    double w2lDist=0.12; // The distance between the center of the wheels and the tilt unit
    
    curlTheta=curwTheta;
    curlXPos=curwXPos+w2lDist*cos(curwTheta);
    curlYPos=curwYPos+w2lDist*sin(curwTheta);
}

