#ifndef LASER_FUSION_H
#define LASER_FUSION_H

#include <stdlib.h>
#include <iostream>
#include <math.h>
#include <vector>
#include <string>
#include <fstream>

#include <libplayerc++/playerc++.h>

//using namespace PlayerCc;

#define INF		4.0950f
#define NUM		655
#define VIW		(229.62267f*(float)M_PI/180.0f)
#define DEV		(VIW/(float)(NUM-1))
#define PI      3.14159265

typedef struct tagPT3D
{
	float x;
	float y;
	float z;
} PT3D;

using namespace std;

class LaserFusion {
    
    public:
//        LaserFusion(PlayerCc::PlayerClient *ap_robot, PlayerCc::Position2dProxy *ap_pp, PlayerCc::PtzProxy *ap_ptz, PlayerCc::LaserProxy *ap_lp);
        LaserFusion(const char *ip_address, int port_num, double tilt_step=5*PI/180, unsigned int tilt_iters=11);

        void updateMeasurements(void);
        
        // The laser data that are accessible by everyone
        vector<double>  m_fusedLaserScan;
        // The position data that are accessible by everyone. The coordinate system is centered on the tilt unit
        double          m_xPos;
        double          m_yPos;
        double          m_theta;
        int             m_laserScanCount; // The number of measurements per scan
        double          m_laserRes; // The resolution of the laser sensor        
        
//        void setTiltAngles(vector<double> &a_tiltAngles);
//        vector<double> getTiltAngles(void);
        
        void setSleepTime(unsigned int a_uSleepTime) { m_uSleepTime=a_uSleepTime; }
        unsigned int getSleepTime(void) { return m_uSleepTime; }
        
    private:
        unsigned int        m_uSleepTime;
        vector<double>      m_tiltAngles; // In radians
        double              m_horizontalOffset; // The horizontal offset of the tilt unit

        // DEBUG VARIABLES
//        int                 m_iteration; // used for debugging
//        string              m_filename;
        
        PlayerCc::PlayerClient      *mp_robot;
        PlayerCc::Position2dProxy   *mp_pp;
        PlayerCc::PtzProxy          *mp_ptz;
        PlayerCc::LaserProxy        *mp_lp;

        // Private functions
        bool TransformRadialTo3D( double fAngle, vector<PT3D>& pts ); // vector<float> OneScan, 
        bool TransformLaserLocalToGlobal( const vector<PT3D>& ptsLocal, double Tx, double Tz, double theta, vector<PT3D>& ptsGlobal );
        bool Quantize( const vector<PT3D>& pts, vector<double>& OneScan );
        void wheelsToLaserCoordinates(double curwXPos, double curwYPos, double curwTheta, double &curlXPos, double &curlYPos, double &curlTheta);
};

#endif
