
#include <stdlib.h>
#include <utils/commandline.h>
#include <gridfastslam/gridslamprocessor.h>
#include <utils/orientedboundingbox.h>
#include <configfile/configfile.h>
#include <libplayerc++/playerc++.h>
#include <sensor/sensor_base/sensorreading.h>
#include <sensor/sensor_range/rangesensor.h>
#include <sensor/sensor_range/rangereading.h>
//#include </usr/local/include/player-2.0/libplayerc++/playerc++.h>
#include <iostream>
#include <time.h>

#include <QtGui/qpainter.h>
#include <QtGui/qpixmap.h>
#include <QtGui/qwidget.h>
#include <QtGui/qwmatrix.h>
#include <Qt/qtextstream.h>
#include <QtGui/qimage.h>


using namespace GMapping;
using namespace std;

int main(int argc, const char * const * argv){

	try
		{
//		double xmin=-10.;//-5.;
//		double ymin=-10.;//-5.;
//		double xmax=10.;//5.;
//		double ymax=10.;//5.;
		double xmin=-5.;
		double ymin=-5.;
		double xmax=5.;
		double ymax=5.;
		double delta=0.05;

		//scan matching parameters
		double sigma=0.05;
		double maxrange=4;
		double maxUrange=4;
		double regscore=1e4;
		double  lstep=.05;
		double astep=.05;
		int kernelSize=1;
		int iterations=5;
		double critscore=0.;
		double maxMove=1.;
		double lsigma=.075;
		double ogain=3;
		int lskip=0;

		//motion model parameters
		double srr=0.01, srt=0.01, str=0.01, stt=0.01;
		//particle parameters
		int particles=30;


		//gfs parameters
		double angularUpdate=0.5;
		double linearUpdate=0.2;
		double resampleThreshold=0.5;
		bool generateMap=true;


		//CREATION

		using namespace PlayerCc;

		PlayerClient robot("192.168.1.102", 6665);
		Position2dProxy pp(&robot, 0);
		LaserProxy lp(&robot, 0);
		std::cout << robot << std::endl;
        robot.SetDataMode(PLAYER_DATAMODE_PULL); // Get data on demand. Ensures that the most up-to-date data are received
        robot.SetReplaceRule(true,-1,-1,-1); // Setup a replace rule, so the queues are not overflowed. Check "playerclient.h" for details
		pp.SetMotorEnable(true);

		GridSlamProcessor* processor=new GridSlamProcessor;
		robot.Read();

		//SENSOR MAP 
		double res= lp.GetScanRes();
		int count = lp.GetCount();

		int loop = 0;
		while( res == 0 )
		{
			robot.Read();
			std::cout << count++ << std::endl;
			res= lp.GetScanRes();
			count = lp.GetCount();
			loop++;
			if( loop > 10 )
				return -1;
		}

		assert(res>0);
		string sensorName="FLASER";
		OrientedPoint rpose(pp.GetXPos(), pp.GetYPos(), pp.GetYaw()); // Doesn't make any sense to me
		OrientedPoint lpose(pp.GetXPos(), pp.GetYPos(), pp.GetYaw());
		OrientedPoint dp=absoluteDifference(lpose, rpose);
		RangeSensor*m_frontLaser=new RangeSensor(sensorName,count, res, OrientedPoint(0,0,0), 0,maxUrange);
		m_frontLaser->updateBeamsLookup(); // Probably it's not needed. The constructor calls this function by default
		SensorMap m_sensorMap;
		m_sensorMap.insert(make_pair(sensorName, m_frontLaser));
			
		processor->setSensorMap(m_sensorMap);

		//set the command line parameters
		processor->setMatchingParameters(maxUrange, maxrange, sigma, kernelSize, lstep, astep, iterations, lsigma, ogain, lskip);
		processor->setMotionModelParameters(srr, srt, str, stt);
		processor->setUpdateDistances(linearUpdate, angularUpdate, resampleThreshold);
		processor->setgenerateMap(generateMap);
		OrientedPoint initialPose((xmin+xmax)/2, (ymin+ymax)/2, 0);
//		OrientedPoint initialPose(xmin+xmax/2, ymin+ymax/2, 0);


		//INITIALIZATION
		processor->init(particles, xmin, ymin, xmax, ymax, delta, initialPose);

		bool running=true;

		GridSlamProcessor* ap, *copy=processor->clone();
		ap=processor; processor=copy; copy=ap;

		//this is the CORE LOOP;	
		// go into read-think-act loop
		int idx = 0;
		srand(time(NULL));
//pp.SetMotorEnable(true);
//pp.SetSpeed(0.2, 0);
		for(;;)
		{
			double newspeed = 0;
			double newturnrate = 0;
			double minR = 1e9;
			double minL = 1e9;

			// this blocks until new data comes; 10Hz by default
			robot.Read();

			// laser avoid (stolen from esben's java example)
			uint count = lp.GetCount();
/*			for (uint j=round(count/4); j < count/2; ++j)
			{
				if (minR > lp[j])
					minR = lp[j];
			}
			for (uint j = count/2; j < round(count*3/4); ++j)
			{
				if (minL > lp[j])
					minL = lp[j];
			}
			std::cout << "minR: " << minR
				<< "minL: " << minL
				<< std::endl;

			double l = (1e5*minR)/500-100;
			double r = (1e5*minL)/500-100;

			bool bTurnLeft = false;
			if( l>r )
				bTurnLeft = true;

			if (l > 100)
				l = 100;
			if (r > 100)
				r = 100;
			
			if( !bTurnLeft )
				l = round(l * 0.95);
			else
				r = round(r * 0.95);
*/
            for (uint j=count/4; j < count/2; ++j)
            {
                if (minR > lp[j])
                    minR = lp[j];
            }
            for (uint j = count/2; j < 3.*count/4; ++j)
            {
                if (minL > lp[j])
                    minL = lp[j];
            }

//            double l = (1e5*minR)/500-100;
//            double r = (1e5*minL)/500-100;
            double l = 200*minL-100;
            double r = 200*minR-100;

            if (l > 100)
                l = 100;
            if (r > 100)
                r = 100;

            std::cout << "minR: " << minR  << "minL: " << minL << "\tl: " << l << "\tr: " << r << std::endl;

            newspeed = (r+l)/(2.*1e2);
//            newspeed = (r+l)/1e3;

            newturnrate = (l-r)/5.;
//            newturnrate = (r-l);
            newturnrate = limit(newturnrate, -40.0, 40.0);
            newturnrate = dtor(newturnrate);
            
            int sleeptime = 100000;
            if (newspeed<0.1){
                if(time(NULL)%2==1)
                    newturnrate=0.5;
                else
                    newturnrate=-0.5;
                

                //newturnrate=0.15;//*rand();
                newspeed = 0;
                sleeptime *= 20;
            }
            
            std::cout << "speed: " << newspeed << "\tturn: " << newturnrate << std::endl;

			// write commands to robot
			pp.SetSpeed(newspeed, newturnrate);

            usleep(sleeptime);
//std::cout << "sleeping.... " << endl;		
//continue;


/*			/////////////////////////////
			/// get readings and set slam
			/////////////////////////////
			const RangeSensor * rs=m_frontLaser;
			RangeReading reading(rs, time(NULL));
			reading.resize(rs->beams().size());
			for (unsigned int i=0; i< (unsigned int)count; i++)
			{
				reading[i]=(double)lp[i];
			}
			reading.setPose(OrientedPoint(pp.GetXPos(), pp.GetYPos(), pp.GetYaw()));


			bool processed=processor->processScan(reading);

			//this returns true when the algorithm effectively processes (the traveled path since the last processing is over a given threshold)
			if (processed){
				//for searching for the BEST PARTICLE INDEX
				unsigned int best_idx=processor->getBestParticleIndex();

				//then if you want to access the BEST MAP,
				//of course by copying it in a plain structure 
				//at this point mymap is yours. Can do what you want.
				Map<double, DoubleArray2D, false>* mymap = processor->getParticles()[best_idx].map.toDoubleMap();

				//dump the map to somewhere.......
				double scale = 50;
				QImage image(int(round((ymax-ymin)*scale*.95)),int(round((xmax-xmin))*scale*.95), QImage::Format_RGB32);
//image.save("tmp.png","png");
				image.fill(qRgb(200,200,255));

				bool bHasSthToDraw = false;
				for(int x=0; x<image.width(); x++)
					for(int y=0; y<image.height(); y++)
					{
						IntPoint tempC = IntPoint(x-image.width()/2, y-image.height()/2);
						Point p=Point(tempC.x/scale, -tempC.y/scale);

						double v=mymap->cell(p);
						if (v>=0)
						{
							int grayValue=255-(int)(255.*v);
							image.setPixel(x,y,qRgb(grayValue, grayValue, grayValue));
							bHasSthToDraw = true;
						}
					}

				if(bHasSthToDraw)
				{
					char name[100];
//					sprintf(name,"map.png", idx++);
					sprintf(name,"map.png");
					image.save(name,"PNG");
//					bool rv=image.save(name,"PNG");
				}

				GridSlamProcessor* newProcessor=processor->clone();
				delete processor;
				processor=newProcessor;
			}
*/			
		}
	}
	catch (PlayerCc::PlayerError e)
	{
		std::cerr << e << std::endl;
		return -1;
	}
	return 0;
}

